#ifndef FILE_TIME_HPP
#define FILE_TIME_HPP

#include <cassert>
#include <windows.h>
#include <hrt/filetime_counter.hpp>
#include <hrt/time_span.hpp>
#include <hrt/time_stamp.hpp>

namespace hrt {

class file_time
{
public:
	//
	// TYPES
	//
	typedef time_stamp<filetime_counter>	time_stamp_type;
	typedef time_span<filetime_counter>		time_span_type;
	typedef filetime_counter::value_type	value_type;

public:
	//
	// CONSTRUCTORS
	//
	file_time()
		: m_nTime(*reinterpret_cast<value_type*>(&m_time))
	{
		m_nTime = 0;
	}

	file_time(const FILETIME* pft)
		: m_nTime(*reinterpret_cast<value_type*>(&m_time))
	{
		assert(NULL != pft);
		m_nTime = *reinterpret_cast<const value_type*>(pft);
	}

	file_time(const file_time& other)
		: m_nTime(*reinterpret_cast<value_type*>(&m_time))
	{
		m_nTime = other.m_nTime;
	}

	file_time(const SYSTEMTIME* pst)
		: m_nTime(*reinterpret_cast<value_type*>(&m_time))
	{
		assert(NULL != pst);
		::SystemTimeToFileTime(pst, &m_time);
	}

	file_time(const value_type&  nTime)
		: m_nTime(*reinterpret_cast<value_type*>(&m_time))
	{
		m_nTime = nTime;
	}

	//
	// ACCESSORS
	//
	operator const FILETIME*() const
		{ return &m_time; }

	bool as_systemtime(LPSYSTEMTIME lpst) const
	{
		assert(NULL != lpst);
		return (0 != FileTimeToSystemTime(&m_time, lpst));
	}

	bool operator==(const file_time& other) const
		{ return (m_nTime == other.m_nTime); }

	bool operator!=(const file_time& other) const
		{ return !(operator==(other)); }

	bool operator>(const file_time& other) const
		{ return m_nTime > other.m_nTime; }

	bool operator<(const file_time& other) const
		{ return m_nTime < other.m_nTime; }

	bool operator>=(const file_time& other) const
		{ return m_nTime >= other.m_nTime; }

	bool operator<=(const file_time& other) const
		{ return m_nTime <= other.m_nTime; }
	
	time_span_type
	operator-(const file_time& other)
		{ return time_stamp_type(m_nTime) - time_stamp_type(other.m_nTime); }

	file_time operator+(const time_span_type& span)
		{ return file_time(m_nTime+span.value()); }

	file_time operator-(const time_span_type& span)
		{ return file_time(m_nTime-span.value()); }

	//
	// MANIPULATORS
	//
	FILETIME* pointer()
	{
		return &m_time;
	}

	file_time& operator=(const file_time& other)
	{
		if (this != &other)
		{
			m_nTime = other.m_nTime;
		}

		return *this;
	}

	file_time& operator+=(const time_span_type& span)
	{
		m_nTime += span.value();
		return *this;
	}

	file_time& operator-=(const time_span_type& span)
	{
		m_nTime -= span.value();
		return *this;
	}

private:
	//
	// IMPLEMENTATION
	//
	FILETIME		m_time;
	value_type&	m_nTime;
};

} // namespace hrt

#endif // FILE_TIME_HPP
