#ifndef FILTERS_HPP
#define FILTERS_HPP

#include <cassert>
#include <algorithm>
#include <stdexcept>
#include <deque>

#ifdef NDEBUG
# pragma warning (push)
# pragma warning (disable : 4702) // get rid of unreachable code warning in release builds
#endif

#include <list>

#ifdef NDEBUG
# pragma warning (pop)
#endif

namespace hrt {

namespace internal {

//
// FILTERING RELATED STUFF
//
class abstract_filter
{
public:
	//
	// CONSTRUCTORS
	//
	virtual ~abstract_filter() {}

	//
	// MANIPULATORS
	//
	double push(double value)
	{ return m_value = on_push(value); }
	void reset()
	{ on_reset(); m_value = 0.0; }

	//
	// ACCESSORS
	//
	virtual double value() const
	{ return m_value; }

protected:
	//
	// CONSTRUCTORS
	//
	abstract_filter(double initial_value)
		: m_value(initial_value)
		{}

	//
	// MORE MANIPULATORS
	//
	virtual double on_push(double value) = 0; // Override and return filtered value
	virtual void on_reset() = 0;							// perform specific reset

private:
	//
	// IMPLEMENTATION
	//
	double m_value;
};

class median_filter : public abstract_filter
{
public:
	//
	// CONSTRUCTORS
	//
	median_filter(int width)
		: abstract_filter(0.0),
			m_values(width, 0.0)
	{	assert(m_values.size()%2 != 0);	}

protected:
	//
	// MANIPULATORS
	//
	double on_push(double value)
	{
		m_values.pop_front();
		m_values.push_back(value);
		std::deque<double> sorted(m_values);
		std::sort(sorted.begin(), sorted.end());
		
		return sorted.at(m_values.size()/2);
	}

	void on_reset()
	{ m_values.assign(m_values.size(), 0.0); }

private:
	//
	// IMPLEMENTATION
	//
	std::deque<double> m_values;
};

class iir_filter : public abstract_filter
{
public:
	//
	// CONSTRUCTORS
	//
	iir_filter(double old_factor)
		: abstract_filter(0.0),
			OLD_FACTOR(old_factor),
			NEW_FACTOR(1.0 - old_factor)
	{
		if (old_factor > 1.0)
			throw std::logic_error("Weighting factor for old value cannot exceed 1.0 (100%)");
	}

	iir_filter(const iir_filter& other)
		: abstract_filter(0.0),
			OLD_FACTOR(other.OLD_FACTOR),
			NEW_FACTOR(other.NEW_FACTOR)
	{}

protected:
	//
	// MANIPULATORS
	//
	double on_push(double value)
	{ return (abstract_filter::value()*OLD_FACTOR)+(value*NEW_FACTOR); }
	void on_reset() {}; 

private:
	iir_filter& operator=(const iir_filter&); // Not implemented

	//
	// IMPLEMENTATION
	//
	const double OLD_FACTOR, NEW_FACTOR;
};

class multi_filter : public abstract_filter
{
public:
	//
	// CONSTRUCTORS
	//
	multi_filter(abstract_filter* p_first)
		: abstract_filter(0.0),
			m_filters(1, p_first)
	{
		if (0 == p_first)
		{
			throw std::invalid_argument("Invalid pointer passed to multi_filter::multi_filter");
		}
	}
	~multi_filter()
	{
		std::for_each<std::list<abstract_filter*>::iterator, void(*)(void*)>(m_filters.begin(), m_filters.end(), (::operator delete));
	}

	//
	// MANIPULATORS
	//
	void add_filter(abstract_filter* p_new)
	{
		if (0 == p_new)
		{
			throw std::invalid_argument("Invalid pointer passed to multi_filter::add_filter");
		}

		m_filters.push_back(p_new);
	}

protected:
	//
	// MANIPULATORS
	//
	double on_push(double value)
	{
		for (std::list<abstract_filter*>::iterator iter = m_filters.begin(); iter != m_filters.end(); iter++)
		{
			value = (*iter)->push(value);
		}

		return value;
	}

	void on_reset()
	{
		for (std::list<abstract_filter*>::iterator iter = m_filters.begin(); iter != m_filters.end(); iter++)
		{
			(*iter)->reset();
		}
	}

private:
	//
	// IMPLEMENTATION
	//
	std::list<abstract_filter*> m_filters;
};

} // namespace internal

} // namespace hrt

#endif // FILTERS_HPP
