#ifndef TIME_STATISTICS_HPP
#define TIME_STATISTICS_HPP
//
// statistics: Class that dispatches timekeeping statistics
//
// Two basic types of statistics are supported:
//
// - Monitor records (monitors drift and errors)
// - Synch records (describes synchs to system time)
//
// To use:
//
// #define TIME_ENABLE_STATISTICS
//
// When this is defined, all monitor and synch data are dispatched 
// to registered listeners (try not to spend an awful lot of time in
// these callbacks).
//

//
// Make null definitions
//
#define TIMESTATS_VAR(var, type, initial)
#define TIMESTATS_STMT(stmt)

#ifdef TIME_ENABLE_STATISTICS

//
// Make real definitions
//
#undef  TIMESTATS_VAR
#define TIMESTATS_VAR(var, type, initial) type var = initial
#undef  TIMESTATS_STMT
#define TIMESTATS_STMT(stmt) stmt

//
// Avoid level-4 warnings from xtree (VC7, 7.1) and vector (VC7.1)
// in release builds (unreachable code).
//
#if defined (_MSC_VER) && _MSC_VER >= 1300
#	ifdef NDEBUG
#		pragma warning (push)
#		pragma warning (disable: 4702)
#	endif
#endif

#include <map>
#include <vector>

#if defined (_MSC_VER) && _MSC_VER >= 1300
#	ifdef NDEBUG
#		pragma warning (pop)
#	endif
#endif

#include <cassert>
#include <utility>
#include <limits>
#include <windows.h>
#include <hrt/filetime_counter.hpp>
#include <hrt/internal/reference_point.hpp>


namespace hrt {

namespace internal {

inline bool is_nan(const double& dbl)
{
	return (!(dbl > 0.0) && !(dbl <= 0.0));
}

inline bool is_infinity(const double& dbl)
{
	return std::numeric_limits<double>::infinity() == dbl;
}

inline double fix_double(const double& dbl)
{
	return ((is_infinity(dbl) || is_nan(dbl)) ? 0.0 : dbl);
}

template<typename counter_type>
class statistics
{
public:
	//
	// CONSTANTS
	//
	enum
	{
		POST_REGISTRATION_HINT,
		DATA_HINT,
		PRE_UNREGISTRATION_HINT
	};

public:
	//
	// TYPES
	//
	typedef filetime_counter::value_type			time_type;
	typedef reference_point<counter_type>			reference_point_type;
	typedef counter_type											counter_type;
	typedef typename counter_type::value_type	counter_value_type;
	typedef int																cb_id_type;

	struct synchronize_stat
	{
		synchronize_stat(int inner_loops_,
										 int outer_loops_,
										 const reference_point_type& ref_point_)
			: inner_loops(inner_loops_),
				outer_loops(outer_loops_),
				ref_point(ref_point_)
		{
			::GetSystemTimeAsFileTime(&creation_time);
		}

		FILETIME creation_time;
		int inner_loops;
		int outer_loops;
		reference_point_type ref_point;
	};

	struct monitor_record
	{
		monitor_record(double seconds_elapsed_,
									 double filtered_diff_,
                   double raw_diff_, 
                   double error_,
                   double frequency_,
                   double filtered_next_,
                   double raw_next_)
		: seconds_elapsed(seconds_elapsed_),
      filtered_diff(filtered_diff_),
      raw_diff(raw_diff_),
      error(error_),
      frequency(frequency_),
      filtered_next(filtered_next_),
      raw_next(raw_next_)
		{
			::GetSystemTimeAsFileTime(&creation_time);
		}

		FILETIME creation_time;
		double seconds_elapsed,
					 filtered_diff,
           raw_diff, 
           error,
           frequency,
           filtered_next,
           raw_next;
	};

	typedef void (*synchronize_cb_type)(const synchronize_stat&, void*, int);
	typedef void (*monitor_cb_type)(const monitor_record&, void*, int);

public:
	//
	// MANIPULATORS
	//
	void push_synchronize(int inner_loops,
												int outer_loops,
												const reference_point_type& rp)
	{
		lock lk(&m_cs);

		synchronize_stat synch_stat(inner_loops, outer_loops, rp);

		safe_dispatch(m_synch_cbs, synch_stat);
	}

	cb_id_type register_synchronize_callback(synchronize_cb_type synch_cb, void* pv)
	{
		return register_callback(m_synch_cbs, synch_cb, pv);
	}

	bool unregister_synchronize_callback(cb_id_type callback_id)
	{
		return unregister_callback(m_synch_cbs, callback_id);
	}

	void push_monitor_record(double seconds_elapsed,
													 double filtered_diff,
													 double raw_diff, 
													 double error,
													 double frequency,
													 double filtered_next,
													 double raw_next)
	{
		monitor_record rec(fix_double(seconds_elapsed),
											 fix_double(filtered_diff),
                       fix_double(raw_diff),
                       fix_double(error),
                       fix_double(frequency),
                       fix_double(filtered_next),
                       fix_double(raw_next));

		
		safe_dispatch(m_monitor_cbs, rec);
	}

	cb_id_type register_monitor_callback(monitor_cb_type monitor_cb, void* pv)
	{
		return register_callback(m_monitor_cbs, monitor_cb, pv);
	}

	bool unregister_monitor_callback(cb_id_type callback_id)
	{
		return unregister_callback(m_monitor_cbs, callback_id);
	}

public:
	//
	// CONSTRUCTORS
	//
	statistics()
	{
		::InitializeCriticalSection(&m_cs);
	}

	~statistics()
	{
		::DeleteCriticalSection(&m_cs);
	}

private:
	//
	// HELPERS
	//
	template<class map_type, class callback_type>
	cb_id_type register_callback(map_type& cb_map, callback_type& cb, void* pv)
	{
		assert(0 != cb);

		lock lk(&m_cs);

		cb_id_type cb_id = std::numeric_limits<cb_id_type>::min();

		while (cb_id < std::numeric_limits<cb_id_type>::max())
		{
			if (cb_map.end() == cb_map.find(cb_id))
			{
				cb_map[cb_id] = std::make_pair(cb, pv);
				break;
			}

			cb_id++;
		}

		assert(cb_id != std::numeric_limits<cb_id_type>::max());

		(*cb)(cb_data_traits<map_type>::empty_value(), pv, POST_REGISTRATION_HINT);

		return cb_id;
	}

	template<class map_type>
	bool unregister_callback(map_type& cb_map, cb_id_type cb_id)
	{
		lock lk(&m_cs);

		map_type::iterator iter = cb_map.find(cb_id);
		if (cb_map.end() != iter)
		{
			const map_type::value_type::second_type& cb_data = (*iter).second;
			(*(cb_data.first))(cb_data_traits<map_type>::empty_value(),
												 cb_data.second,
                         PRE_UNREGISTRATION_HINT);
		}

		return (0 != cb_map.erase(cb_id));
	}

	template<class map_type, class cb_data_type>
  safe_dispatch(map_type& cb_map, const cb_data_type& cb_data)
	{
		typedef std::vector<map_type::key_type>	cb_id_vector;

		lock lk(&m_cs);

		cb_id_vector failed_cbs(cb_map.size());

		for (map_type::const_iterator iter = cb_map.begin();
				 iter != cb_map.end();
				 ++iter)
		{
			try
			{
				(*((*iter).second.first))(cb_data, ((*iter).second.second), DATA_HINT);
			}
			catch (...)
			{
				failed_cbs.push_back((*iter).first);
			}
		}

		for (cb_id_vector::const_iterator iter = failed_cbs.begin();
				 iter != failed_cbs.end();
				 ++iter)
		{
			cb_map.erase(*iter);
		}
	}

private:
	//
	// INTERNAL TYPES
	//
	struct lock
	{
		lock(LPCRITICAL_SECTION pcs) : m_pcs(pcs)
			{ assert(0 != m_pcs); ::EnterCriticalSection(m_pcs); }

		~lock()
      { ::LeaveCriticalSection(m_pcs); }

	private:
		LPCRITICAL_SECTION m_pcs;
	};

private:
	//
	// PRIVATE TYPES
	//
	typedef std::map<cb_id_type, std::pair<synchronize_cb_type, void*> > synch_cb_map;
	typedef std::map<cb_id_type, std::pair<monitor_cb_type, void*> > monitor_cb_map;

	template<typename cb_map_type>
	struct cb_data_traits
	{
		static void empty_value(); // needs specialization
	};

	template<> struct cb_data_traits<synch_cb_map>
	{
		static synchronize_stat empty_value()
		{ return synchronize_stat(0, 0, reference_point_type(0,0,0,0)); }
	};

	template<> struct cb_data_traits<monitor_cb_map>
	{
		static monitor_record empty_value()
		{ return monitor_record(0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0); }
	};

private:
	//
	// DATA
	//
	synch_cb_map			m_synch_cbs;
	monitor_cb_map		m_monitor_cbs;
	CRITICAL_SECTION	m_cs;
};

} // namespace internal

} // namespace hrt

#endif // TIME_ENABLE_STATISTICS

#endif // TIME_STATISTICS_HPP
