#ifndef SYNCHRONIZE_HPP
#define SYNCHRONIZE_HPP

//
// Avoid level-4 warnings from vector (unreachable code) in VC 7.1
// in release builds
//
#if defined (_MSC_VER) && _MSC_VER >= 1310
#	ifdef NDEBUG
#		pragma warning (push)
#		pragma warning (disable: 4702)
#	endif
#endif

#include <vector>

#if defined (_MSC_VER) && _MSC_VER >= 1310
#	ifdef NDEBUG
#		pragma warning (pop)
#	endif
#endif

#include <algorithm>
#include <hrt/internal/reference_point.hpp>
#include <hrt/internal/statistics.hpp>
#include <hrt/internal/scoped_priority_class.hpp>
#include <hrt/internal/scoped_thread_priority.hpp>
#include <hrt/internal/thread_priority.hpp>

namespace hrt {

namespace internal {

#pragma warning (push)
#pragma warning (disable : 4127) // conditional expression is constant
#pragma warning (disable : 4390) // empty control statement

#pragma optimize("", off)
#if defined (TIME_ENABLE_STATISTICS)

template<typename continuous_ticker_t>
typename continuous_ticker_t::value_type
synchronize(const typename continuous_ticker_t::value_type& desired_accuracy,
						reference_point<continuous_ticker_t>& ref_point,
						statistics<continuous_ticker_t>* p_stats = 0,
						int failures_before_increasing_priority = 2)

#else

template<typename continuous_ticker_t>
typename continuous_ticker_t::value_type
synchronize(const typename continuous_ticker_t::value_type& desired_accuracy,
						reference_point<continuous_ticker_t>& ref_point,
						int failures_before_increasing_priority = 2)

#endif // TIME_ENABLE_STATISTICS
{
	typedef typename continuous_ticker_t::value_type	continuous_value;
	typedef typename filetime_counter::value_type			filetime_value;

	const continuous_value LIMIT = desired_accuracy * 2; // See post-loop for explanation
	continuous_value continuous0 = 0,
									 continuous1 = 0,
									 prev_diff = 0;
	filetime_value	 filetime0 = 0,
									 filetime1 = 0;
	thread_priority	 current_priority;
	int							 outer_loops = 0;

	TIMESTATS_VAR(inner_loops, int, 0);

	//
	// Synchronize
	//
	do
	{
		TIMESTATS_STMT(outer_loops++);
		if (0 == (outer_loops % failures_before_increasing_priority))
		{
			current_priority++;
		}

		prev_diff = 0;
		continuous0 = continuous_ticker_t::value();
		filetime0 = filetime_counter::value();
		
		do
		{
			TIMESTATS_STMT(inner_loops++);
			filetime1 = filetime_counter::value();
			continuous1 = continuous_ticker_t::value();
			if (filetime0 != filetime1)
				break;
			prev_diff = continuous1 - continuous0;
			continuous0 = continuous1;
		}
		while (true);

	}
	while ((!filetime_counter::is_tickincrement(filetime1 - filetime0)) ||	// make sure that we didn't miss a clock cycle
				 ((continuous1 - continuous0 + prev_diff) >= LIMIT));							// allow LIMIT difference

	//
	// Adjust counter value to middle of interval (we will then always
	// have at least desired_accuracy precision)
	//
	continuous_value guaranteed_accuracy = static_cast<continuous_value>((double(continuous1 - continuous0 + prev_diff) / 2.0) + 0.5);
	ref_point.set_counter(continuous1 - guaranteed_accuracy);
	ref_point.set_filetime(filetime1);
	ref_point.set_accuracy(guaranteed_accuracy);
	ref_point.set_tickincrement(filetime1 - filetime0);

	TIMESTATS_STMT(p_stats ? p_stats->push_synchronize(inner_loops,
																										 outer_loops,
                                                     ref_point) : (void)0 );

	return guaranteed_accuracy;
}

template<typename continuous_ticker_t>
typename continuous_ticker_t::value_type
determine_synch_accuracy()
{
	typedef typename continuous_ticker_t::value_type	continuous_value;

	const int															NO_OF_EXECUTIONS = 25;
	const continuous_value								unused_accuracy = continuous_ticker_t::frequency()/1000; // 1ms
	reference_point<continuous_ticker_t>	unused_ref_pt(0,0,0);

	std::vector<continuous_value>	execution_times;

	execution_times.reserve(NO_OF_EXECUTIONS);

	scoped_priority_class		spc(REALTIME_PRIORITY_CLASS);
	scoped_thread_priority	stp(THREAD_PRIORITY_TIME_CRITICAL);

	for (int i = 0; i < NO_OF_EXECUTIONS; ++i)
	{
		execution_times.push_back(synchronize<continuous_ticker_t>(unused_accuracy, unused_ref_pt));
	}

	//
	// Return median time
	//
	std::partial_sort(execution_times.begin(), execution_times.begin()+(NO_OF_EXECUTIONS/2), execution_times.end());

	return execution_times[(NO_OF_EXECUTIONS/2)-1];
}
#pragma optimize( "", on )
#pragma warning (pop) // restore warnings

} // namespace internal

} // namespace hrt

#endif // SYNCHRONIZE_HPP
