#ifndef SYNCHRONIZER_HPP
#define SYNCHRONIZER_HPP

#include <algorithm>
#include <cassert>
#include <stdexcept>
#include <windows.h>
#include <hrt/internal/synchronize.hpp>
#include <hrt/internal/time_data.hpp>
#include <hrt/internal/statistics.hpp>
#include <hrt/internal/helpers.hpp>
#include <hrt/internal/filters.hpp>
#include <hrt/internal/reference_point.hpp>
#include <hrt/internal/scoped_thread_priority.hpp>
#include <hrt/internal/scoped_thread_affinity.hpp>
#include <hrt/filetime_counter.hpp>
#include <hrt/time_span.hpp>
#include <hrt/time_stamp.hpp>
#include <hrt/timer.hpp>

#if (_WIN32_WINNT < 0x0400)
# error _WIN32_WINNT must be defined as 0x0400 or better (waitable timer support)
#endif

namespace hrt {

namespace internal {

//
// template class synchronizer: The guts of the time provider (in dire need of refactoring).
//
template<typename	counter_type>
class synchronizer
{
private:
	//
	// PRIVATE TYPES
	//
	typedef timer<counter_type>										timer_type;
	typedef reference_point<counter_type>					reference_point_type;
	typedef time_data<counter_type>								time_data;
	typedef typename timer_type::time_stamp_type	time_stamp_type;
	typedef typename timer_type::time_span_type		time_span_type;
	typedef typename filetime_counter::value_type	time_type;
	typedef typename counter_type::value_type			counter_value_type;

	typedef synchronizer<counter_type> this_type;

public:
	//
	// CONSTRUCTORS
	//
	synchronizer(int keep_within_micros,
							 int synchronize_thread_priority,
							 double tuning_limit,
							 int max_wait_millis,
							 int min_wait_millis,
							 DWORD_PTR affinity_mask = USE_AUTO_AFFINITY)
		: m_synch_limit(micros_to_secs(keep_within_micros)*0.75),
			m_synchronize_thread_priority(synchronize_thread_priority),
			m_tuning_limit(tuning_limit),
			m_max_wait(millis_to_secs(max_wait_millis)),
			m_min_wait(millis_to_secs(min_wait_millis)),
			m_tdInUse(0),
			m_ptd(&m_td[0]),
			m_filter(new median_filter(3)), // FIX: Hardcoded value
			m_quit_event(::CreateEvent(NULL, TRUE, FALSE, NULL)),
			m_synch_accuracy(0),
			m_tick_check_timer(::CreateWaitableTimer(NULL, FALSE, NULL)),
      m_tune_timer(::CreateWaitableTimer(NULL, TRUE, NULL)),
			m_affinity_mask(USE_AUTO_AFFINITY == affinity_mask ? best_synchronize_affinity() : affinity_mask)
	{
		m_filter.add_filter(new iir_filter(0.6));  // FIX: Hardcoded value

		validate_affinity_mask();

		//
		// Check stop event
		//
		if (NULL == m_quit_event)
		{
			throw std::runtime_error("Failed to create event");
		}

		//
		// Check timers
		//
		if ((NULL == m_tick_check_timer) || (NULL == m_tune_timer))
		{
			throw std::runtime_error("Failed to create timers");
		}

		//
		// Determine appropriate synch accuracy
		//
		scoped_thread_affinity affinity(m_affinity_mask);
		m_synch_accuracy = determine_synch_accuracy<counter_type>();
		affinity.dismiss();

		//
		// Set initial data
		//
		double initial_frequency = static_cast<double>(m_timer.frequency());
		if (counter_traits<counter_type>::needs_initial_adjustment())
		{
			DWORD time_adjustment = 0,
						time_increment = 0;
			BOOL	adjustment_disabled = TRUE;

			if ((0 != ::GetSystemTimeAdjustment(&time_adjustment, &time_increment, &adjustment_disabled)) &&
					(FALSE == adjustment_disabled))
			{
				initial_frequency *= static_cast<double>(time_increment)/static_cast<double>(time_adjustment);
			}
		}
		reset(initial_frequency, get_reference());

		//
		// Kick off our monitoring thread
		//
		HANDLE hTuner = ::CreateThread(NULL, 0, &synchronizer_thread, this, 0, &m_dwTunerId);
		assert(NULL != hTuner);
		if (NULL == hTuner)
		{
			throw std::runtime_error("Failed to start tuner thread");
		}
		::CloseHandle(hTuner);
	}

	~synchronizer()
	{
		//
		// Implementation note: We deliberately do NOT use the thread handle
		// when waiting for the tuner thread to terminate. We also do not
		// wait for the event to be signalled if the call to PTM fails,
		// as this most probably means that the thread has already ended.
		//
		if (0 != ::PostThreadMessage(m_dwTunerId, WM_QUIT, 0, 0))
		{
			::WaitForSingleObject(m_quit_event, INFINITE);
		}
		::CloseHandle(m_quit_event);
		::CloseHandle(m_tick_check_timer);
		::CloseHandle(m_tune_timer);
	}

public:
	//
	// PUBLIC TYPES
	//
	typedef counter_type											counter_type;
	typedef typename counter_type::value_type	raw_value_type;
	TIMESTATS_STMT(typedef statistics<counter_type>	statistics_type);

public:
	//
	// ACCESSORS
	//
	inline raw_value_type rawtime() const
	{
		return m_timer.now().value();
	}

	inline time_data timedata() const
	{
		return time_data(*front_buffer());
	}

#ifdef TIME_ENABLE_STATISTICS
	statistics_type& stats() const
	{
		return m_stats;
	}
#endif

private:
	//
	// Tuner thread
	//
	static DWORD WINAPI synchronizer_thread(LPVOID p)
	{
		this_type*	pthis = (this_type*)p;
		TCHAR				szWndClass[255];

		wsprintfA(szWndClass, "synchronizer_window:%lu:%p", GetTickCount(), pthis);

		WNDCLASS wc =
		{
			0,											// style
			&synchronizer_wndproc,	// WNDPROC
			0,											// class-extra
			0,											// window-extra
			GetModuleHandle(NULL),	// instance
			NULL,										// no icon
			NULL,										// no cursor
			NULL,										// no brush
			NULL,										// no menu
			szWndClass							// window class
		};

		HANDLE hTunerWnd = ::CreateWindowA((LPCSTR)::RegisterClassA(&wc),
																			 "synchronizer_window",
																				0,
																				CW_USEDEFAULT,
																				CW_USEDEFAULT,
																				CW_USEDEFAULT,
																				CW_USEDEFAULT,
																				NULL,
																				NULL,
																				::GetModuleHandle(NULL),
																				pthis);

		assert(NULL != hTunerWnd);

		DWORD result = pthis->run();

		::SetEvent(pthis->m_quit_event);
		::CloseHandle(hTunerWnd);

		return result;
	}

	static LRESULT CALLBACK synchronizer_wndproc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
	{
		return ::DefWindowProc(hWnd, uMsg, wParam, lParam);
	}

	DWORD run()
	{
		DWORD		dwWaitResult = 0;
		bool		fContinue = true;
		HANDLE	timer_handles[2] = { tick_check_timer(), tune_timer() };

		scoped_thread_affinity affinity(m_affinity_mask);

		//
		// Initial setup of timers
		//
		set_periodic_timer(tick_check_timer(), millis_to_secs(TICKINCREMENT_CHECK_INTERVAL_MILLIS));
		set_oneshot_timer(tune_timer(), default_wait());

		do
		{
			//
			// Wait for message or timeout
			//
			dwWaitResult = ::MsgWaitForMultipleObjects(sizeof(timer_handles)/sizeof(timer_handles[0]),
																								 &timer_handles[0],
																								 FALSE,
																								 TICKINCREMENT_CHECK_INTERVAL_MILLIS,
																								 QS_ALLINPUT);

			switch (dwWaitResult)
			{
				case (WAIT_OBJECT_0):
				{
					//
					// Tick check timer expired, check if tick period changed
					//
					if (last_ref().tickincrement() != filetime_counter::tickincrement())
					{
						double next_tune = on_tickincrement_changed();
						set_oneshot_timer(tune_timer(), next_tune);
					}
					break;
				}
	
				case (WAIT_OBJECT_0+1):
				{
					//
					// Tune timer expired, time to check drift
					//
					double next_tune = on_tune();
					set_oneshot_timer(tune_timer(), next_tune);
					break;
				}

				case (WAIT_OBJECT_0+2):
				{
					//
					// Message(s) available
					//
					MSG msg;

					while (FALSE != ::PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
					{
						switch (msg.message)
						{
							case WM_QUIT:
							{
								fContinue = false;
								break;
							}
							case WM_TIMECHANGE:
							{
								double next_tune = on_time_changed();
								set_oneshot_timer(tune_timer(), next_tune);
								break;
							}
							default:
							{
								::TranslateMessage(&msg);
								::DispatchMessage(&msg);
							}
						}		// switch msg.message
					}			// while FALSE != PeekMessage(...)

					break;
				}

				default:
				{
					assert(!"Unexpected wait result");
					fContinue = false;
					break;
				}
			}
		} while (fContinue);

		return ::GetLastError();
	}

private:
	//
	// HELPERS
	//
	reference_point_type get_reference() const
	{
		scoped_thread_priority stp(m_synchronize_thread_priority);

		reference_point_type ref_pt;
#if defined (TIME_ENABLE_STATISTICS)
		::hrt::internal::synchronize<counter_type>(m_synch_accuracy, ref_pt, &m_stats);
#else
		::hrt::internal::synchronize<counter_type>(m_synch_accuracy, ref_pt);
#endif

		return ref_pt;
	}

	void reset(double frequency, const reference_point_type& ref_pt)
	{
		time_data* ptd = back_buffer();

		ptd->set_ref(ref_pt);
		ptd->set_frequency(frequency);

		m_filter.reset();
		m_last_ref_pt = ref_pt;

		flip();
	}

	bool time_changed_recently() const
	{
		MSG msg;
		while (FALSE != ::PeekMessage(&msg, NULL, 0, 0, PM_REMOVE))
		{
			switch (msg.message)
			{
				case WM_TIMECHANGE:
				{
					//
					// Ok, so it might have changed
					//
					return true;
				}

				case WM_QUIT:
				{
					//
					// Repost and return immediately
					//
					::PostThreadMessage(GetCurrentThreadId(), WM_QUIT, 0, 0);

					return false;
				}

				default:
				{
					::TranslateMessage(&msg);
					::DispatchMessage(&msg);
					break;
				}
			}
		}

		return false;
	}

	double on_tune()
	{
		double											next_tune = default_wait();
		reference_point_type				ref_pt = get_reference();
		const reference_point_type& last_ref_pt = last_ref();
		const double								frequency = current_frequency();

		//
		// We need to check these two explicitly in case something changed during
		// get_reference.
		//
		if (time_changed_recently() || (last_ref_pt.tickincrement() != ref_pt.tickincrement()))
		{
			apply_timesetting_changed(ref_pt);
			return default_wait();
		}

		//
		// OK, looking good - now check if we need to resync
		//
		time_span_type ts = time_stamp_type(ref_pt.counter()) - last_ref_pt.counter();
		const double seconds_elapsed_calculated = ts.value()/frequency;
		const double seconds_elapsed_real = double(ref_pt.filetime() - last_ref_pt.filetime())/(double)filetime_counter::frequency();

		if (0.0 > seconds_elapsed_calculated)
		{
			//
			// Oops, counter did wrap
			//
			reset(frequency, ref_pt);
			return default_wait();
		}

		double raw_diff = seconds_elapsed_calculated - seconds_elapsed_real;
		m_filter.push(raw_diff);
		double error = (m_filter.value()/seconds_elapsed_real);

		if (drift_exceeds_limit(m_filter.value()))
		{
			TIMESTATS_STMT(m_stats.push_monitor_record(seconds_elapsed_real,
																								 m_filter.value(),
                                                 raw_diff,
                                                 error,
                                                 frequency,
                                                 0.0,
                                                 0.0));

			reset(adjusted_frequency_from_error(error), ref_pt);
		}
		else
		{
			//
			// Try to determine worst (soonest) case for next threshold crossing
			//
			double raw_next = (synch_limit() - abs(raw_diff)) / (abs(raw_diff/seconds_elapsed_real));
			raw_next = std::max(min_wait(), raw_next);
			double next = (error == 0.0) ? 0.0 : (synch_limit() - abs(m_filter.value()))/abs(error);
			next = std::max(min_wait(), next);

			TIMESTATS_STMT(m_stats.push_monitor_record(seconds_elapsed_real,
																								 m_filter.value(),
                                                 raw_diff,
                                                 error,
                                                 frequency,
                                                 next,
                                                 raw_next));

			double worst_case_tune = std::min(raw_next, next);
			if (worst_case_tune < next_tune)
				next_tune = worst_case_tune;
			else
				next_tune = std::min(worst_case_tune, max_wait());
		}

		return next_tune;
	}

	double on_time_changed()
	{
		apply_timesetting_changed(get_reference());
		return default_wait();
	}

	double on_tickincrement_changed()
	{
		apply_timesetting_changed(get_reference());
		return default_wait();
	}

	void apply_timesetting_changed(const reference_point_type& new_ref)
	{
		double new_frequency = current_frequency();

		if (new_ref.tickincrement() != last_ref().tickincrement())
		{
			double correction_factor = static_cast<double>(last_ref().tickincrement())/static_cast<double>(new_ref.tickincrement());
			new_frequency *= correction_factor;
		}

		reset(new_frequency, new_ref);
	}

	void validate_affinity_mask() const
	{
		DWORD_PTR process_affinity = 0,
							system_affinity = 0;

		if ((0 == ::GetProcessAffinityMask(::GetCurrentProcess(), &process_affinity, &system_affinity)) ||
				(DWORD_PTR(0) == (process_affinity & m_affinity_mask)))
		{
			throw std::runtime_error("Invalid affinity mask specified for synchronization thread");
		}
	}

	double adjusted_frequency_from_error(double error) const
	{
		return current_frequency()*(1.0+normalize_error(error));
	}

	double normalize_error(double error) const
	{
		double abs_error = abs(error);

		if (abs_error < min_correction())
		{
			return 0.0;
		}

		abs_error = std::min(abs_error, max_correction());

		return (error < 0.0) ? -abs_error : abs_error;
	}

private:
	//
	// Access to timing data (fast and 'almost' thread-safe)
	//
	inline const time_data* front_buffer() const
	{
		time_data* ptd;
		InterlockedExchangePointer(reinterpret_cast<void**>(&ptd), m_ptd);
		assert(0 != ptd);
		return ptd;
	}

	inline time_data* back_buffer()
	{
		return &m_td[!m_tdInUse];
	}

	inline void flip()
	{
		InterlockedExchangePointer(reinterpret_cast<void**>(&m_ptd), back_buffer());
		m_tdInUse = !m_tdInUse;
	}

private:
	//
	// HELPERS
	//
	double default_wait() const
		{ return millis_to_secs(DEFAULT_TUNE_INTERVAL_MILLIS); }

	double min_wait() const
		{ return m_min_wait; }

	double max_wait() const
		{ return m_max_wait; }

	double synch_limit() const
		{ return m_synch_limit; }

	bool drift_exceeds_limit(const double& drift) const
		{ return !within_bounds(-synch_limit(), drift, synch_limit()); }

	HANDLE tick_check_timer() const
		{ return m_tick_check_timer; }

	HANDLE tune_timer() const
		{ return m_tune_timer; }

	const reference_point_type& last_ref() const
		{ return m_last_ref_pt; }

	double current_frequency() const
		{ return front_buffer()->frequency(); }

	double max_correction() const
		{ return MAX_CORRECTION_PPM/1000000.0; }

	double min_correction() const
		{ return m_tuning_limit; }

private:
	//
	// PRIVATE STATIC CONSTANTS
	//
	enum
	{
		TICKINCREMENT_CHECK_INTERVAL_MILLIS = 100,
		DEFAULT_TUNE_INTERVAL_MILLIS = 500,
		MILLIS_PER_SECOND = 1000,
		MICROS_PER_SECOND = 1000000,
		USE_AUTO_AFFINITY = 0,
		MAX_CORRECTION_PPM = 10
	};

private:
	//
	// STATIC HELPERS
	//
	template<typename int_type, typename float_type>
	static int_type float_to_int(const float_type& source)
		{ return static_cast<int_type>(source+0.5); }

	static double millis_to_secs(int millis)
		{ return static_cast<double>(millis) / MILLIS_PER_SECOND; }

	static double micros_to_secs(int micros)
		{ return static_cast<double>(micros) / MICROS_PER_SECOND; }

	static int secs_to_millis(const double& secs)
		{ return float_to_int<int>(secs * MILLIS_PER_SECOND); }

	static __int64 secs_to_relativetime(const double& secs)
		{ return -(float_to_int<__int64>(secs*filetime_counter::frequency())); }

	static bool set_periodic_timer(HANDLE timer, double period)
	{
		assert(NULL != timer);
		__int64 initial_timeout = secs_to_relativetime(period);
		return FALSE != ::SetWaitableTimer(timer,
																			 reinterpret_cast<const LARGE_INTEGER*>(&initial_timeout),
																			 secs_to_millis(period),
																			 NULL,
																			 NULL,
																			 FALSE);
	}

	static bool set_oneshot_timer(HANDLE timer, double timeout)
	{
		assert(NULL != timer);
		__int64 initial_timeout = secs_to_relativetime(timeout);
		return FALSE != ::SetWaitableTimer(timer,
																			 reinterpret_cast<const LARGE_INTEGER*>(&initial_timeout),
																			 0,
																			 NULL,
																			 NULL,
																			 FALSE);
	}

	static DWORD_PTR best_synchronize_affinity()
	{
		DWORD_PTR	process_affinity_mask = 0,
							system_affinity_mask = 0,
							synchronize_mask = 0x01;

		::GetProcessAffinityMask(::GetCurrentProcess(), &process_affinity_mask, &system_affinity_mask);
		if (0x01 != process_affinity_mask)
		{
			//
			// We're on an SMP system and allowed to run on other CPUs than #0 - do it!
			//
			synchronize_mask = (process_affinity_mask & ~DWORD_PTR(0x01));
		}

		return synchronize_mask;
	}

private:
	//
	// Non-copyable
	//
	this_type& operator=(const this_type&);	// Not implemented
	this_type(const this_type&);						// -"-

private:
	//
	// IMPLEMENTATION (all internal times in seconds as floating-point data)
	//
	// (I'm really ashamed of all these member variables ... must refactor ... soon.)
	//
	timer_type	m_timer;			// Timer source
	double			m_synch_limit;
	int					m_synchronize_thread_priority;
	double			m_tuning_limit;
	double			m_max_wait,
							m_min_wait;
	HANDLE			m_quit_event;		// Used for signalling that tuner thread received WM_QUIT
	DWORD				m_dwTunerId;	// -"-
	time_data		m_td[2];			// Flip-flop buffer (read/write)
	int					m_tdInUse;		// Index of current time data
	time_data*	m_ptd;				// Pointer to current time data
	multi_filter	m_filter;			// Smoothening (less data-spike sensitive) filter for time deviations
	typename counter_type::value_type	m_synch_accuracy; // accuracy to use for synchs
	HANDLE			m_tick_check_timer,	// Timer for checking tick period
							m_tune_timer;				// Timer for waiting for next tune
	reference_point_type	m_last_ref_pt;	// Reference at last synch
	DWORD_PTR		m_affinity_mask;

#ifdef TIME_ENABLE_STATISTICS
	mutable statistics_type m_stats;
#endif
};

} // namespace internal

} // namespace hrt

#endif SYNCHRONIZER_HPP
