#ifndef SYSTEM_TIME_HPP
#define SYSTEM_TIME_HPP

#include <cassert>
#include <stdexcept>
#include <windows.h>
#include <hrt/file_time.hpp>

namespace hrt {

class system_time
{
public:
	//
	// CONSTRUCTORS
	//
	system_time()
		{ clear(); }
	system_time(const SYSTEMTIME& time)
		: m_time(time)
		{}
	system_time(const FILETIME* pft)
	{
		clear();
		::FileTimeToSystemTime(pft, &m_time);
	}

	system_time(WORD wYear,
							WORD wMonth,
							WORD wDay,
							WORD wHour,
							WORD wMinute,
							WORD wSecond,
							WORD wMillis)
	{
		FILETIME ft;
		m_time.wYear = wYear;
		m_time.wMonth = wMonth;
		m_time.wDay = wDay;
		m_time.wHour = wHour;
		m_time.wMinute = wMinute;
		m_time.wSecond = wSecond;
		m_time.wMilliseconds = wMillis;
		m_time.wDayOfWeek = 0;

		//
		// Just a quirk to get verify and get day of week
		//
		if (SystemTimeToFileTime(&m_time, &ft))
			FileTimeToSystemTime(&ft, &m_time);
		else
		{
			throw std::invalid_argument("system_time::system_time: invalid parameters");
		}
	}

	//
	// ACCESSORS
	//
	int year() const
		{ return m_time.wYear; }
	int month() const
		{	return m_time.wMonth; };
	int day() const
		{	return m_time.wDay; };
	int day_of_week() const
		{	return m_time.wDayOfWeek; };
	int hour() const
		{ return m_time.wHour; };
	int minute() const
		{ return m_time.wMinute; };
	int second() const
		{ return m_time.wSecond; };
	int millis() const
		{ return m_time.wMilliseconds; };
	operator const SYSTEMTIME*() const
		{ return &m_time; }
	bool operator==(const system_time& other) const
	{
		bool fRetVal = true;

		if (this != &other)
		{
			fRetVal = (file_time(&m_time) == file_time(&other.m_time));
		}

		return fRetVal;
	}

	bool operator!=(const system_time& other) const
	{
		return !(operator==(other));
	}

	bool operator>(const system_time& other) const
	{
		return (file_time(&m_time) > file_time(&other.m_time));
	}

	bool operator<(const system_time& other) const
	{
		return (file_time(&m_time) < file_time(&other.m_time));
	}

	bool operator>=(const system_time& other) const
	{
		return (file_time(&m_time) >= file_time(&other.m_time));
	}

	bool operator<=(const system_time& other) const
	{
		return (file_time(&m_time) <= file_time(&other.m_time));
	}

	//
	// MANIPULATORS (Note: almost no validation)
	//
	void set_year(int year)
		{ assert(year <= (unsigned short)-1); m_time.wYear = static_cast<WORD>(year); }
	void set_month(int month)
		{	assert((month >= 1) && (month <= 12)); m_time.wMonth = static_cast<WORD>(month); };
	void set_day(int day)
		{	assert((day >= 1) && (day <= 31)); m_time.wDay = static_cast<WORD>(day); };
	void set_day_of_week(int dow)
		{	assert((dow >= 0) && (dow <= 6)); m_time.wDayOfWeek = static_cast<WORD>(dow); };
	void set_hour(int hour)
		{ assert((hour >= 0) && (hour <= 24)); m_time.wHour = static_cast<WORD>(hour); };
	void set_minute(int minute)
		{ assert((minute >= 0) && (minute <= 59)); m_time.wMinute = static_cast<WORD>(minute); };
	void set_second(int second)
		{ assert((second >= 0) && (second <= 59)); m_time.wSecond = static_cast<WORD>(second); };
	void set_millis(int millis)
		{ assert((millis >= 0) && (millis <= 999)); m_time.wMilliseconds = static_cast<WORD>(millis); };

	void as_filetime(FILETIME* pft) const
	{
		assert(NULL != pft);
		::SystemTimeToFileTime(&m_time, pft);
	}

	//
	// MANIPULATORS
	//
	LPSYSTEMTIME pointer()
		{ return &m_time; }
	
	system_time& operator=(const SYSTEMTIME& time)
	{
		m_time = time;
		return *this;
	}

private:
	//
	// HELPERS
	//
	void clear()
	{
		::ZeroMemory(&m_time, sizeof(SYSTEMTIME));
	}

	SYSTEMTIME m_time;
};

} // namespace hrt

#endif // SYSTEM_TIME_HPP
