#ifndef TIME_PROVIDER_HPP
#define TIME_PROVIDER_HPP

#include <cassert>
#include <windows.h>
#include <hrt/filetime_counter.hpp>
#include <hrt/time_span.hpp>
#include <hrt/time_stamp.hpp>
#include <hrt/internal/synchronizer.hpp>
#include <hrt/internal/time_data.hpp>

namespace hrt {

//
// Class time_provider: provides a high resolution absolute time
//
template<
				 typename	counter_type,
				 int KEEP_WITHIN_MICROS	=	100,
				 int SYNCHRONIZE_THREAD_PRIORITY = THREAD_PRIORITY_BELOW_NORMAL,
				 int TUNING_LIMIT_PARTSPERBILLION	=	100,	// Don't tune	if error less	than 0.1 ppm
				 int MAX_WAIT_MILLIS = 10000,
				 int MIN_WAIT_MILLIS = 10
				 >
class time_provider
{
private:
	//
	// PRIVATE TYPES
	//
	typedef internal::synchronizer<counter_type>	synchronizer_type;
	typedef internal::time_data<counter_type>			time_data_type;
	typedef time_stamp<counter_type>							time_stamp_type;
	typedef time_span<counter_type>								time_span_type;
	typedef filetime_counter::value_type					time_type;

private:
	//
	// CONSTRUCTORS
	//
	time_provider()
		: m_synchronizer(KEEP_WITHIN_MICROS,
										 SYNCHRONIZE_THREAD_PRIORITY,
										 TUNING_LIMIT_PARTSPERBILLION/1000000000.0,
										 MAX_WAIT_MILLIS,
										 MIN_WAIT_MILLIS)
	{}

	time_provider& operator=(const time_provider&);	// Not implemented
	time_provider(const time_provider&);						// -"-

public:
	//
	// PUBLIC TYPES
	//
#ifdef TIME_ENABLE_STATISTICS
	typedef typename synchronizer_type::statistics_type	statistics_type;
#endif

	typedef counter_type												counter_type;
	typedef typename counter_type::value_type		raw_value_type;

public:
	//
	// CONSTRUCTORS
	//
	static time_provider& instance()
	{
		static time_provider the_instance;
		return the_instance;
	}

public:
	//
	// ACCESSORS
	//
	inline bool systemtime(LPSYSTEMTIME p_system_time) const
	{
		assert(0 != p_system_time);
		bool			retval = false;
		FILETIME	ft = { 0, 0 };

		if (filetime(&ft) && ::FileTimeToSystemTime(&ft, p_system_time))
		{
			retval = true;
		}

		return retval;
	}

	inline bool filetime(LPFILETIME p_file_time) const
	{
		assert(0 != p_file_time);
		return filetime_from_rawtime(rawtime(), p_file_time);
	}

	inline raw_value_type rawtime() const
	{
		return m_synchronizer.rawtime();
	}

	inline bool systemtime_from_rawtime(const raw_value_type& raw_value, LPSYSTEMTIME p_system_time) const
	{
		assert(0 != p_system_time);
		bool			retval = false;
		FILETIME	ft = { 0, 0 };

		if (filetime_from_rawtime(raw_value, &ft) &&
				::FileTimeToSystemTime(&ft, p_system_time))
		{
			retval = true;
		}

		return retval;
	}

	inline bool filetime_from_rawtime(const raw_value_type& raw_value, LPFILETIME p_file_time) const
	{
		assert(0 != p_file_time);
		time_data_type	td(m_synchronizer.timedata());
		time_span_type	elapsed = time_stamp_type(raw_value) - td.ref().counter();

		time_type ft_ticks = (time_type)(double(elapsed.value())/td.relative_frequency());
		time_type ft_time = td.ref().filetime() + ft_ticks;

		*p_file_time = *(reinterpret_cast<LPFILETIME>(&ft_time));

		return true;
	}

#ifdef TIME_ENABLE_STATISTICS
	statistics_type& stats() const
	{
		return m_synchronizer.stats();
	}
#endif

private:
	//
	// IMPLEMENTATION
	//
	synchronizer_type	m_synchronizer;
};

} // namespace hrt

#endif // TIMEPROVIDER_HPP
