#ifndef TIME_SPAN_HPP
#define TIME_SPAN_HPP

namespace hrt {

//
// Class time_span: represents a high-resolution time span
//
template<typename counter_type>
class time_span
{
public:
	//
	// TYPES
	//
	typedef counter_type											counter_type;
	typedef typename counter_type::value_type	value_type;

public:
	//
	// CONSTRUCTORS
	//
	time_span()
		: m_span(0),
			m_frequency(counter_type::frequency()),
			m_dirty(true)
		{}
	time_span(const value_type& span)
		: m_span(span),
			m_frequency(counter_type::frequency()),
			m_dirty(true)
		{}
	time_span(const time_span& other)
		: m_span(other.m_span),
			m_frequency(counter_type::frequency()),
			m_dirty(true)
		{}
	time_span(int days, int hours, int minutes, int seconds, value_type nanos = 0i64)
		: m_span(0),
			m_frequency(counter_type::frequency()),
			m_dirty(true)
	{
		_ASSERTE((0 <= hours) && (23 >= hours));
		_ASSERTE((0 <= minutes) && (59 >= minutes));
		_ASSERTE((0 <= seconds) && (59 >= seconds));
		_ASSERTE((0 <= nanos) && (999999999i64 >= nanos));

		m_span += (value_type)SECONDS_PER_DAY*(value_type)days;
		m_span += (value_type)SECONDS_PER_HOUR*(value_type)hours;
		m_span += (value_type)SECONDS_PER_MINUTE*(value_type)minutes;
		m_span += (value_type)seconds;
		m_span *= m_frequency;

		//
		// Add nanoseconds
		//
		m_span += (m_frequency*(value_type)nanos)/1000000000i64;
	}

	//
	// MANIPULATORS
	//
	time_span& operator+=(const time_span& other)
		{ m_span += other.m_span; set_dirty(); return *this; }
	time_span& operator-=(const time_span& other)
		{ m_span -= other.m_span; set_dirty(); return *this; }
	time_span& operator=(const time_span& other)
		{ if (this != &other) { m_span = other.m_span; set_dirty(); } return *this; }

	//
	// ACCESSORS
	//
	bool operator==(const time_span& other)
		{	return m_span == other.m_span; }
	bool operator!=(const time_span& other)
		{	return m_span != other.m_span; }
	bool operator>(const time_span& other)
		{	return m_span > other.m_span; }
	bool operator<(const time_span& other)
		{	return m_span < other.m_span; }
	bool operator>=(const time_span& other)
		{	return m_span >= other.m_span; }
	bool operator<=(const time_span& other)
		{	return m_span <= other.m_span; }

	time_span operator+(const time_span& other) const
		{ return time_span(m_span + other.m_span); }
	time_span operator-(const time_span& other) const
		{ return time_span(m_span - other.m_span); }

	value_type value() const
		{ return m_span; }
	double total_seconds() const
		{ return (double)m_span/(double)m_frequency; }

	//
	// ACCESSORS
	//
	int days() const
		{	update(); return m_split.m_days; };
	int hours() const
		{ update(); return m_split.m_hours; };
	int minutes() const
		{ update(); return m_split.m_minutes; };
	int seconds() const
		{ update(); return m_split.m_seconds; };
	int millis() const
		{ update(); return m_split.m_millis; };
	int micros() const
		{ update(); return m_split.m_micros; };
	int nanos() const
		{ update(); return m_split.m_nanos; };

private:
	//
	// CONSTANTS
	//
	enum
	{
		SECONDS_PER_MINUTE = 60,
		SECONDS_PER_HOUR = 60*60,
		SECONDS_PER_DAY = 60*60*24
	};

	//
	// HELPERS
	//
	void update() const
	{
		if (m_dirty)
		{
			value_type temp = m_span;

			m_split.m_days = static_cast<int>(temp/(SECONDS_PER_DAY*m_frequency));
			temp -= m_split.m_days*SECONDS_PER_DAY*m_frequency;
			m_split.m_hours = static_cast<int>(temp/(SECONDS_PER_HOUR*m_frequency));
			temp -= m_split.m_hours*SECONDS_PER_HOUR*m_frequency;
			m_split.m_minutes = static_cast<int>(temp/(SECONDS_PER_MINUTE*m_frequency));
			temp -= m_split.m_minutes*SECONDS_PER_MINUTE*m_frequency;
			m_split.m_seconds = static_cast<int>(temp/m_frequency);
			temp -= m_split.m_seconds*m_frequency;
			m_split.m_millis = static_cast<int>((1000i64*temp)/m_frequency);
			m_split.m_micros = static_cast<int>((1000000i64*temp)/m_frequency);
			m_split.m_nanos = static_cast<int>((1000000000i64*temp)/m_frequency);

			set_dirty(false);
		}
	}

	void set_dirty(bool flag = true) const
	{
		m_dirty = flag;
	}

	//
	// IMPLEMENTATION
	//
	value_type		m_span;
	value_type		m_frequency;
	mutable bool	m_dirty;
	mutable struct _info
	{
		int m_days;
		int m_hours;
		int m_minutes;
		int m_seconds;
		int m_millis;
		int m_micros;
		int m_nanos;

		//
		// Constructor just used to zero out
		//
		_info()
			: m_days(0),
				m_hours(0),
				m_minutes(0),
				m_seconds(0),
				m_millis(0),
				m_micros(0),
				m_nanos(0)
			{}
	}
	m_split;
};

} // namespace hrt

#endif // TIME_SPAN_HPP
