#ifndef TSC_COUNTER_HPP
#define TSC_COUNTER_HPP

#include <hrt/counter_traits.hpp>
#include <hrt/filetime_counter.hpp>
#include <hrt/performance_counter.hpp>
#include <hrt/internal/scoped_priority_class.hpp>
#include <hrt/internal/scoped_thread_priority.hpp>
#include <windows.h>
#include <cassert>
#include <stdexcept>
#include <string>

namespace hrt 
{

struct use_safe_tsc_raw_value {};
struct use_fast_tsc_raw_value {};

template<typename raw_value_mode = use_safe_tsc_raw_value>
struct tsc_counter
{
public:
	//
	// TYPES
	//
	typedef __int64													value_type;
	typedef filetime_counter::value_type		filetime_value;
	typedef performance_counter::value_type	performance_value;

private:
	//
	// HELPERS
	//
	inline static void get_times(filetime_value& filetime, value_type& tsc)
	{
		LARGE_INTEGER	li;
		unsigned			base = 0,
									base_extra = 0;

		__asm
		{
			;
			; warm-up as cpuid takes longer time
			; first few executions
			;
			cpuid
			rdtsc
			mov	li.HighPart, edx
			mov	li.LowPart, eax
			cpuid
			rdtsc
			mov	li.HighPart, edx
			mov	li.LowPart, eax

			;
			; real-stuff
			;
			cpuid
			rdtsc
			mov	li.HighPart, edx
			mov	li.LowPart, eax
			cpuid
			rdtsc
			sub eax, li.LowPart
			mov base_extra, eax

			cpuid
			rdtsc
			mov	li.HighPart, edx
			mov	li.LowPart, eax
			cpuid
			rdtsc
			sub eax, li.LowPart
			mov base, eax
		}

		//
		// To be safe for rollover
		//
		if (base_extra < base)
			base = base_extra;

		filetime_value		t0,t1;
		performance_value	limit = performance_counter::frequency()/200000,
											counter0 = 0i64,
											counter1 = 0i64;

restart:

		//
		// Synchronize
		//
		do
		{
			counter1 = performance_counter::value();
			t0 = filetime_counter::value();
			
			do
			{
				counter0 = performance_counter::value();
				if ((counter0 - counter1) > limit)
				{
					goto restart;
				}
				t1 = filetime_counter::value();
				__asm
				{
					cpuid
					rdtsc
					mov	li.HighPart, edx
					mov	li.LowPart, eax
				}
				counter1 = performance_counter::value();
			}
			while (t0 == t1);

		}
		while ((!filetime_counter::is_tickincrement(t1 - t0)) ||	// make sure that we didn't miss a clock cycle
					 ((counter1 - counter0) > limit));	// allow only <limit> difference

		filetime = t1;
		tsc = li.QuadPart;
		tsc -= base;

		base_time() = base;
	}

	static bool test_hardware()
	{
		//
		// test for rdtsc instruction
		//
		__try
		{
			__asm rdtsc
		}
		__except (EXCEPTION_EXECUTE_HANDLER)
		{
			return false;
		}

		//
		// test for possible no-op
		//
		unsigned long low1 = 0,
									low2 = 0;
		__asm
		{
			cpuid
			rdtsc
			mov	low1, eax
			cpuid
			rdtsc
			mov	low2, eax
		}

		return (low1 != low2);
	}

	static value_type coarse_calibrate()
	{
		//
		// Test hardware
		//
		if (!test_hardware())
			throw std::runtime_error("Hardware does not support the rdtsc instruction");

		//
		// Test platform
		//
		OSVERSIONINFO ovi;
		::ZeroMemory(&ovi, sizeof(ovi));
		ovi.dwOSVersionInfoSize = sizeof(ovi);

		if ((FALSE != GetVersionEx(&ovi)) &&
				(ovi.dwMajorVersion <= 5) &&
				(ovi.dwMinorVersion <= 1) &&
				(1195000i64 < performance_counter::frequency()))
		{
			//
			// Assume that the system's using the pentium counter as
			// the qpc_counter. It's probably closer to the truth than
			// our coarse calibration.
			//
			return static_cast<value_type>(performance_counter::frequency());
		}

		int						current_measurement = 0;
		const double	ALLOWED_DEVIATION = 0.001; // 0.1%
		const int			SAMPLE_ARRAY_SIZE = 10;
		value_type		measurements[SAMPLE_ARRAY_SIZE],
									frequency = 0;
		bool					calibrated = false;

		for (int i = 0; i < SAMPLE_ARRAY_SIZE; measurements[i++] = 0i64);

		//
		// Raise priority to avoid chance of getting re-scheduled while
		// performing synchronization.
		//
		internal::scoped_priority_class		spc(REALTIME_PRIORITY_CLASS);
		internal::scoped_thread_priority	stp(THREAD_PRIORITY_TIME_CRITICAL);

		while (!calibrated)
		{
			filetime_value	t0 = 0,
											t1 = 0;
			value_type			q0 = 0,
											q1 = 0;

			get_times(t0, q0);
			::Sleep(50);
			get_times(t1, q1);
			assert(t1 > t0);

			measurements[current_measurement] = (10000000i64*(q1-q0)/(t1-t0));

			//
			// If the latest measurements are equal within
			// specified tolerance, return frequency as the
			// average of these.
			//
			value_type	nMin = measurements[0],
									nMax = nMin,
                  nTotal = 0;

			for (int j = 0; j < SAMPLE_ARRAY_SIZE; j++)
			{
				if (measurements[j] < nMin)
					nMin = measurements[j];
				if (measurements[j] > nMax)
					nMax = measurements[j];

				nTotal += measurements[j];
			}

			double average = (double)nTotal/(double)SAMPLE_ARRAY_SIZE;
			double error = ((double)(nMax-nMin)/average);

			if (error <= ALLOWED_DEVIATION)
			{
				frequency = static_cast<value_type>(average+0.5);
				calibrated = true;
			}

			current_measurement++;
			current_measurement %= SAMPLE_ARRAY_SIZE;
		}

		return frequency;
	}

	static inline value_type& base_time()
	{
		static value_type the_base = 0;
		return the_base;
	}

public:

	static inline value_type frequency()
	{
		static value_type frequency = coarse_calibrate();
		return frequency;
	}

	static value_type raw_value();

	static inline value_type value()
	{
		return raw_value() - base_time();
	}
};

#pragma warning(push)
#pragma warning(disable:4035) // avoid 'no return value' warning from 'raw_value'.

template<>
inline
tsc_counter<use_fast_tsc_raw_value>::value_type
tsc_counter<use_fast_tsc_raw_value>::raw_value()
{ 
	__asm
	{
		rdtsc
	};
}

template<>
inline
tsc_counter<use_safe_tsc_raw_value>::value_type
tsc_counter<use_safe_tsc_raw_value>::raw_value()
{ 
	__asm
	{
		cpuid
		rdtsc
	};
}

#pragma warning(pop)

//
// counter_traits specializations
//
template<>
struct counter_traits<tsc_counter<use_safe_tsc_raw_value> >
{
	static std::string description()
	{ return "time stamp counter (safe)"; }
	static bool needs_initial_adjustment()
	{ return false; }
};

template<>
struct counter_traits<tsc_counter<use_fast_tsc_raw_value> >
{
	static std::string description()
	{ return "time stamp counter (fast)"; }
	static bool needs_initial_adjustment()
	{ return false; }
};

} // namespace hrt

#endif // TSC_COUNTER_HPP
