//
// collstat: Demonstrates collecting time provider statistics.
//
#include "stdafx.h"
#include <cassert>
#include <exception>
#include <iostream>
#include <iomanip>
#include <sstream>
#include <fstream>

using namespace hrt;

//
// Function declarations
//
std::string CreateTimeTaggedFilename(const std::string& prefix, const system_time& st);
void SynchronizeCallback(const synch_stat_t& synchStat, void* pv, int hint);
void MonitorCallback(const monitor_record_t& monitorRecord, void* pv, int hint);
BOOL WINAPI CtrlHandler(DWORD dwCtrlType);

//
// Global quit event
//
HANDLE g_quitEvent = CreateEvent(NULL, TRUE, FALSE, NULL);

int main(int /*argc*/, char* /*argv*/[])
{
	time_provider_t&	provider = time_provider_t::instance();
	system_time				systemTime;

	provider.systemtime(systemTime.pointer());

	//
	// Create and open uniquely named data files
	//
	std::string synchDataFilename = CreateTimeTaggedFilename("synch_data_", systemTime),
							monitorDataFilename = CreateTimeTaggedFilename("monitor_data_", systemTime);

	std::ofstream synchData(synchDataFilename.c_str(), std::ios_base::trunc);
	std::ofstream monitorData(monitorDataFilename.c_str(), std::ios_base::trunc);

	if ((!synchData) || (!monitorData))
	{
		std::cerr << "ERROR: Failed to create one or both of the data files (" << synchDataFilename << ", "
							<< monitorDataFilename << ")." << std::endl;
		return 1;
	}

	//
	// Setup abort routine
	//
	if (FALSE == ::SetConsoleCtrlHandler(&CtrlHandler, TRUE))
	{
		std::cerr << "ERROR: Failed to register control handler." << std::endl;
		return 1;
	}

	//
	// Register listeners
	//
	statistics_t&							stats = provider.stats();
	statistics_t::cb_id_type	synchCbId = stats.register_synchronize_callback(&SynchronizeCallback, &synchData);
	statistics_t::cb_id_type	monitorCbId = stats.register_monitor_callback(&MonitorCallback, &monitorData);

	//
	// Wait until aborted
	//
	std::cout << "Collecting data - <ctrl-c> to stop ..." << std::flush;
	::WaitForSingleObject(g_quitEvent, INFINITE);

	//
	// Unregister & clean up
	//
	stats.unregister_monitor_callback(monitorCbId);
	stats.unregister_synchronize_callback(synchCbId);

	::SetConsoleCtrlHandler(&CtrlHandler, FALSE);
	::CloseHandle(g_quitEvent);

	std::cout << " data available in:\n\n"
						<< "'" << monitorDataFilename << "' (monitor data)\n"
						<< "'" << synchDataFilename << "' (synchronization data)\n"
						<< std::endl;

	return 0;
}


std::string CreateTimeTaggedFilename(const std::string& prefix, const system_time& st)
{
	char								szComputerName[MAX_COMPUTERNAME_LENGTH + 1] = { 0 };
	DWORD								dwBufSize = sizeof(szComputerName)/sizeof(szComputerName[0]);
	std::ostringstream	ostr;

	::GetComputerNameA(szComputerName, &dwBufSize);

	ostr.fill('0');

	ostr << szComputerName
			 << '_'
			 << prefix
			 << std::setw(4) << st.year()
			 << std::setw(2) << st.month()
			 << std::setw(2) << st.day()
			 << '_'
			 << std::setw(2) << st.hour()
			 << std::setw(2) << st.minute()
			 << std::setw(2) << st.second()
			 << std::setw(3) << st.millis()
			 << ".log";

	return ostr.str();
}

//
// Callback function implementation
//
void SynchronizeCallback(const synch_stat_t& ss, void* pv, int hint)
{
	assert(0 != pv);

	std::basic_ostream<char>* pOstream = static_cast<std::basic_ostream<char>*>(pv);

	if (statistics_t::DATA_HINT == hint)
	{
		//
		// Format & write record
		//
		(*pOstream) << file_time(&ss.creation_time) << ','
								<< ss.outer_loops << ','
                << ss.inner_loops << ','
								<< ss.ref_point.counter() << ','
								<< ss.ref_point.filetime() << ','
								<< file_time(ss.ref_point.filetime()) << ','
								<< ss.ref_point.accuracy() << ','
                << ss.ref_point.tickincrement()
								<< '\n'; 
	}
	else if (statistics_t::POST_REGISTRATION_HINT == hint)
	{
		//
		// Write header
		//
		(*pOstream) << "Time,"
								<< "Outer loops,"
								<< "Inner loops,"
								<< "Counter,"
								<< "Filetime (counter),"
								<< "Filetime,"
								<< "Accuracy (counter ticks),"
								<< "Tick increment"
								<< '\n';
	}
	else if (statistics_t::PRE_UNREGISTRATION_HINT == hint)
	{
		//
		// Flush stream
		//
		pOstream->flush();
	}
}

void MonitorCallback(const monitor_record_t& mr, void* pv, int hint)
{
	assert(0 != pv);

	std::basic_ostream<char>* pOstream = static_cast<std::basic_ostream<char>*>(pv);

	if (statistics_t::DATA_HINT == hint)
	{
		//
		// Format & write record
		//
		(*pOstream) << file_time(&mr.creation_time) << ','
								<< std::fixed << std::setprecision(2) << mr.seconds_elapsed << ','
								<< std::fixed << std::setprecision(9) << mr.filtered_diff << ','
								<< std::fixed << std::setprecision(9) << mr.raw_diff << ','
								<< std::fixed << std::setprecision(9) << mr.error << ','
								<< std::fixed << std::setprecision(0) << mr.frequency << ','
								<< std::fixed << std::setprecision(9) << mr.filtered_next << ','
								<< std::fixed << std::setprecision(9) << mr.raw_next
								<< '\n';
	}
	else if (statistics_t::POST_REGISTRATION_HINT == hint)
	{
		//
		// Write header
		//
		(*pOstream) << "Time,"
								<< "Elapsed(s),"
                << "Filtered diff(s),"
                << "Raw diff(s),"
                << "Error,"
                << "Frequency(Hz),"
								<< "Filtered next(s),"
								<< "Raw next(s)"
                << '\n';
	}
	else if (statistics_t::PRE_UNREGISTRATION_HINT == hint)
	{
		//
		// Flush stream
		//
		pOstream->flush();
	}
}

//
// Abort function implementation
//
BOOL WINAPI CtrlHandler(DWORD /*dwCtrlType*/)
{
	if (WAIT_OBJECT_0 != ::WaitForSingleObject(g_quitEvent, 0))
	{
		::SetEvent(g_quitEvent);	
		return TRUE;
	}

	return FALSE;
}
