//
// threads.cpp : Demonstrates using time_provider::raw_time to timestamp messages
//							 originating from different threads.
//
#include "stdafx.h"					// sample_common.h
#include "LoggingThread.h"
#include "OutputLogEvent.h"
#include "SimpleEvent.h"
#include "LoggingImpl.h"
#include <algorithm>
#include <cassert>
#include <iostream>
#include <sstream>
#include <string>

//
// Types
//
typedef LoggingImpl<time_provider_t>		LoggingImplType;
typedef OutputLogEvent<time_provider_t>	OutputLogEventType;

//
// Globals for the lazy programmer
//
HANDLE g_hQuitEvent = ::CreateEvent(NULL, TRUE, FALSE, NULL);

//
// Function prototypes
//
void EnterToContinue(const std::string& sMessageToShow);
BOOL WINAPI ControlHandler(DWORD);
void CloseHandleWrapper(HANDLE);
bool InitLoggingTimer(HANDLE hTimer, int nIntervalMillis, PHANDLE pSignalEvent);

//
// Constants
//
const int SECONDS_OUTPUT_PRECISION = 4;
const int NO_OF_THREADS = 10;
const int OUTPUT_INTERVAL_MILLIS = 500;
const int SIGNAL_INTERVAL_MILLIS = 2000;


int main(int /*argc*/, char* /*argv*/[])
{
	typedef std::vector<HANDLE>	HandleVector;

	time_provider_t&	provider = time_provider_t::instance();
	HandleVector			vThreadHandles;
	LoggingImplType		logging(&provider);
	HANDLE						hTimer = ::CreateWaitableTimer(NULL, FALSE, NULL),
										hSignalEvent = ::CreateEvent(NULL, TRUE, FALSE, NULL);

	EnterToContinue("This application demonstrates the usage of high-resolution\n"
									"time-stamping in the context of multiple threads.\n\n"
									"Pressing <enter> starts, <ctrl-c> ends application ...");

	//
	// Setup control handler
	//
	if (FALSE == ::SetConsoleCtrlHandler(&ControlHandler, TRUE))
	{
		std::cerr << "Failed to setup control handler" << std::endl;
		return EXIT_FAILURE;
	}

	//
	// Init timer
	//
	logging.LogEvent("Initializing logging timer");
	if (!InitLoggingTimer(hTimer, SIGNAL_INTERVAL_MILLIS, &hSignalEvent))
	{
		std::cerr << "Failed to set logging timer" << std::endl;
		return EXIT_FAILURE;
	}

	//
	// Create all threads
	//
	logging.LogEvent("Creating logging threads");
	for (int i = 0; i < NO_OF_THREADS; ++i)
	{
		vThreadHandles.push_back(::CreateLoggingThread(&logging, hSignalEvent, g_hQuitEvent));
	}

	//
	// Loop until quit signalled, output new events periodically
	//
	logging.LogEvent("Entering main loop");
	while (WAIT_OBJECT_0 != ::WaitForMultipleObjectsEx(1, &g_hQuitEvent, TRUE, OUTPUT_INTERVAL_MILLIS, TRUE))
	{
		//
		// Get events
		//
		logging.LogEvent("Woke up, collecting events");
		LoggingImplType::EventVector events;
		logging.GetEvents(events);

		//
		// Output events
		//
		std::ostringstream ostr;
		std::for_each(events.begin(), events.end(), OutputLogEventType(&ostr, &provider, SECONDS_OUTPUT_PRECISION));
		std::cout << ostr.str() << std::flush;
	}

	//
	// Wait for all threads to terminate
	//
	logging.LogEvent("Waiting for threads to terminate");
	::WaitForMultipleObjects(vThreadHandles.size(),
													 &vThreadHandles[0],
													 TRUE,
													 INFINITE);

	//
	// Cleanup
	//
	std::for_each(vThreadHandles.begin(), vThreadHandles.end(), std::ptr_fun(&CloseHandleWrapper));
	::CancelWaitableTimer(hTimer);
	::CloseHandle(hSignalEvent);

	EnterToContinue("Exiting (press <enter> to end program) ...");

	return EXIT_SUCCESS;
}

void EnterToContinue(const std::string& sMessageToShow)
{
	std::string sUnused;
	std::cout << sMessageToShow << std::endl;
	std::cin.putback(0);
	std::cin >> sUnused;
}

//
// ControlHandler: Just set quit event and unregister self
//
BOOL WINAPI ControlHandler(DWORD /*dwControlType*/)
{
	::SetEvent(g_hQuitEvent);
	::SetConsoleCtrlHandler(&ControlHandler, FALSE);
	return TRUE;
}

//
// Simple wrapper to go from __stdcall to __cdecl calling convention
//
void CloseHandleWrapper(HANDLE h)
{
	::CloseHandle(h);
}

//
// Logging timer helpers
//
void CALLBACK LoggingTimerAPC(LPVOID lpSignalEvent,
															DWORD /*dwTimerLowValue*/,
                              DWORD /*dwTimerHighValue*/)
{
	assert(0 != lpSignalEvent);
	::PulseEvent(*static_cast<PHANDLE>(lpSignalEvent));
}

bool InitLoggingTimer(HANDLE hTimer, int nIntervalMillis, PHANDLE pSignalEvent)
{
	assert(0 != pSignalEvent);
	__int64 i64InitialTimeout = -(nIntervalMillis*10000);
	LARGE_INTEGER liInitialTimeout;

	liInitialTimeout.QuadPart = i64InitialTimeout;

	return (0 != ::SetWaitableTimer(hTimer,
																	&liInitialTimeout,
																	nIntervalMillis,
																	&LoggingTimerAPC,
																	pSignalEvent,
																	FALSE));
}

